#include <stdbool.h>
#include <string.h>

#include "lnm/http/consts.h"
#include "lnm/http/loop.h"
#include "lnm/http/loop_internal.h"
#include "lnm/http/req.h"
#include "lnm/loop.h"
#include "lnm/loop_internal.h"

/* static const lnm_http_loop_state lnm_http_loop_state_first_req =
 * lnm_http_loop_state_parse_req; */
static const lnm_http_loop_state lnm_http_loop_state_first_res =
    lnm_http_loop_state_write_headers;

void lnm_http_loop_process_parse_req(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;

  lnm_http_parse_err res = lnm_http_req_parse(
      &ctx->req, &conn->r.buf[conn->r.read], conn->r.size - conn->r.read);

  switch (res) {
  case lnm_http_parse_err_ok:
    conn->r.read += ctx->req.len;
    ctx->state = lnm_http_loop_state_route;
    break;
  case lnm_http_parse_err_incomplete:
    // If the request is already the size of the read buffer, we close the
    // request. Otherwise, we wait for anything read
    if (conn->r.size - conn->r.read == LNM_LOOP_BUF_SIZE) {
      conn->state = lnm_loop_state_end;
    }
    break;
  case lnm_http_parse_err_invalid:
    conn->state = lnm_loop_state_end;
    break;
  case lnm_http_parse_err_unknown_method:
    ctx->res.status = lnm_http_status_method_not_implemented;
    ctx->state = lnm_http_loop_state_first_res;
    break;
  }
}

void lnm_http_loop_process_route(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lnm_http_loop_gctx *gctx = ctx->g;

  // 0: no match
  // 1: matched route, but not method
  // 2: fully matched route
  int match_level = 0;
  lnm_http_route *route;

  for (size_t i = 0; i < gctx->routes.len && match_level < 2; i++) {
    route = gctx->routes.arr[i];
    bool matched_path;

    switch (route->type) {
    case lnm_http_route_type_literal:
      matched_path =
          strncmp(route->route.s, ctx->req.path.s, ctx->req.path.len) == 0;
      break;
    case lnm_http_route_type_regex:
      matched_path =
          regexec(route->route.regex, ctx->req.path.s,
                  LNM_HTTP_MAX_REGEX_GROUPS, ctx->req.path.groups, 0) == 0;
    }

    // Remember the previous match levels
    int new_match_level = matched_path + (route->method == ctx->req.method);
    match_level = match_level < new_match_level ? new_match_level : match_level;
  }

  switch (match_level) {
  case 0:
    ctx->res.status = lnm_http_status_not_found;
    ctx->state = lnm_http_loop_state_first_res;
    break;
  case 1:
    ctx->res.status = lnm_http_status_method_not_allowed;
    ctx->state = lnm_http_loop_state_first_res;
    break;
  case 2:
    ctx->route = route;
    ctx->cur_step = route->step;
    ctx->state = lnm_http_loop_state_parse_headers;
    break;
  }
}

void lnm_http_loop_process_parse_headers(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;

  // TODO

  ctx->state = lnm_http_loop_state_steps;
}

void lnm_http_loop_process_steps(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;
  lnm_http_step *step;

  do {
    step = ctx->cur_step;

    switch (step->fn(conn)) {
    case lnm_http_step_err_done:
      ctx->cur_step = ctx->cur_step->next;
      break;
    case lnm_http_step_err_io_needed:
      break;
    case lnm_http_step_err_close:
      conn->state = lnm_loop_state_end;
      break;
    case lnm_http_step_err_res:
      ctx->state = lnm_http_loop_state_first_res;
      break;
    }
  } while ((step != ctx->cur_step) && (ctx->cur_step != NULL));

  if (ctx->cur_step == NULL) {
    ctx->state = lnm_http_loop_state_write_headers;
  }
}

void (*process_fns[])(lnm_http_conn *conn) = {
    lnm_http_loop_process_parse_req,
    lnm_http_loop_process_route,
    lnm_http_loop_process_parse_headers,
    lnm_http_loop_process_steps,
};

void lnm_http_loop_process(lnm_http_conn *conn) {
  lnm_http_loop_ctx *ctx = conn->ctx;

  lnm_http_loop_state http_loop_state;
  lnm_loop_state loop_state;

  // We stop processing if:
  // - the event loop state has changed, as we need to switch to the other I/O
  // loop
  // - the process fn returned without changing the HTTP loop state, indicating
  // it's waiting for I/O
  do {
    http_loop_state = ctx->state;
    loop_state = conn->state;

    process_fns[http_loop_state](conn);
  } while ((conn->state == loop_state) && (http_loop_state != ctx->state));
}
